#include "image.h"
#include "rle.h"

Image::Image(bool texture)
{
    isTexture=texture;
    texWidth=texHeight=-1;
    img=NULL;
}

Image::Image(int width, int height)
{
    this->texWidth=this->texHeight=-1;
    img=(ColorRGBA *)malloc(width*height*sizeof(ColorRGBA));
    if (img)
    {
        this->texWidth=width;
        this->texHeight=height;
    }
}

Image::~Image(void)
{
    if (img)
        free((void *)img);
}

Image *Image::copyImage(int x, int y, int w, int h)
{
    ColorRGBA t;
    int c,d;
    Image *tmp;
    tmp=new Image(false);
    tmp->create(w,h,(ColorRGBA){0,0,0,0});

    for (d=0;d<h;d++)
        for (c=0;c<w;c++)
        {
            t=getPixel(x+c,y+d);
            tmp->setPixel(c,d,t);
        }
    return tmp;
}

Image *Image::duplicate(void)
{
    return copyImage(0,0,realWidth,realHeight);
}

void Image::clearRegion(int x, int y, int width, int height)
{
    int c,d;
    for (c=0;c<width;c++)
        for (d=0;d<height;d++)
            setPixel(x+c,y+d,(ColorRGBA){0,0,0,0});
}

ColorRGBA *Image::getPointer(void)
{
    return img;
}

int Image::getTextureWidth(void)
{
    return texWidth;
}

int Image::getTextureHeight(void)
{
    return texHeight;
}

ColorRGBA Image::getPixel(int x, int y)
{
    ColorRGBA none = {0,0,0,0};
    if (x>=0 && y>=0 && x<texWidth && y<texHeight)
        return img[y*texWidth+x];
    else
        return none;
}

void Image::setPixel(int x, int y, ColorRGBA color)
{
    if (x>=0 && y>=0 && x<texWidth && y<texHeight)
        img[y*texWidth+x]=color;
}

int Image::getWidth(void)
{
    return realWidth;
}

int Image::getHeight(void)
{
    return realHeight;
}

bool Image::create(int width, int height, ColorRGBA background)
{
    realWidth=width;
    realHeight=height;

    if (isTexture)
    {
        texWidth=texHeight=64;
        while (texWidth<realWidth)
            texWidth*=2;
        while (texHeight<realHeight)
            texHeight*=2;
    }
    else
    {
        texWidth=realWidth;
        texHeight=realHeight;
    }

    img=(ColorRGBA *)malloc(texWidth*texHeight*sizeof(ColorRGBA));
    if (!img)
        return false;

    int x,y;
    for (x=0;x<texWidth;x++)
        for (y=0;y<texHeight;y++)
        {
            int d=y*texWidth+x;
            if (x<width && y<height)
            {
                img[d].red=background.red;
                img[d].green=background.green;
                img[d].blue=background.blue;
                img[d].alpha=background.alpha;
            }
            else
                img[d].red=img[d].green=img[d].blue=img[d].alpha=0;
        }
    return true;
}



bool Image::loadPPM(char *filename)
{
    bool ret=false;
    char textBuffer[256];
    int x,y;
    unsigned char *tmp;
    FILE *fp;
    fp=fopen(filename,"rb");

    if (fp)
    {
        fgets(textBuffer,256,fp);
        if (textBuffer[0]!='P' || textBuffer[1]!='6')
        {
            printf("[Image::loadPMM] Error: not a PMM file\n");
        }
        else
        {
            fgets(textBuffer,256,fp);
            while (textBuffer[0]=='#')
                fgets(textBuffer,256,fp);

            sscanf(textBuffer,"%i %i",&realWidth, &realHeight);
            if (isTexture)
            {
                texWidth=texHeight=64;
                while (texWidth<realWidth)
                    texWidth*=2;
                while (texHeight<realHeight)
                    texHeight*=2;
            }
            else
            {
                texWidth=realWidth;
                texHeight=realHeight;
            }

            fgets(textBuffer,256,fp);
            if (textBuffer[0]!='2' || textBuffer[1]!='5' || textBuffer[2]!='5')
            {
                printf("[Image::loadPMM] Error: Wrong color parameter\n");
            }
            else
            {
                if (img)
                    free((void *)img);
                img=(ColorRGBA *)malloc(texWidth*texHeight*sizeof(ColorRGBA));
                tmp=(unsigned char *)malloc(realWidth*realHeight*3);
                fread(tmp,3,realWidth*realHeight,fp);

                for (y=0;y<texHeight;y++)
                for (x=0;x<texWidth;x++)
                {
                    int d=y*texWidth+x;
                    if (x<realWidth && y<realHeight)
                    {
                        int c=y*realWidth+x;
                        img[d].red=tmp[c*3];
                        img[d].green=tmp[c*3+1];
                        img[d].blue=tmp[c*3+2];
                        img[d].alpha=(tmp[c*3]==0 && tmp[c*3+1]==0 && tmp[c*3+2]== 0 ? 0 : 255);
                    }
                    else
                        img[d].red=img[d].green=img[d].blue=img[d].alpha=0;

                }

                free((void *)tmp);
                ret=true;
            }

        }
        fclose(fp);
    }
    else
    {
        printf("[Image::loadPPM] Error: %s not found\n",filename);
    }
    return true;
}

bool Image::loadRGBA(char *filename)
{
    bool ret=false;
    bool compressed=false;
    char textBuffer[256];
    int x,y,t;
    unsigned char *tmp, *tmp2;
    FILE *fp;
    fp=fopen(filename,"rb");

    if (fp)
    {
        fgets(textBuffer,256,fp);
        if (textBuffer[0]!='r' || textBuffer[1]!='g' || textBuffer[2]!='b' || textBuffer[3]!='a')
        {
            printf("[Image::loadRGBA] Error: not a RGBA file\n");
        }
        else
        {
            compressed=(textBuffer[4]=='!');
            fgets(textBuffer,256,fp);

            sscanf(textBuffer,"%i %i",&realWidth, &realHeight);
            if (isTexture)
            {
                texWidth=texHeight=64;
                while (texWidth<realWidth)
                    texWidth*=2;
                while (texHeight<realHeight)
                    texHeight*=2;
            }
            else
            {
                texWidth=realWidth;
                texHeight=realHeight;
            }

            if (img)
                free((void *)img);

            img=(ColorRGBA *)malloc(texWidth*texHeight*sizeof(ColorRGBA));
            tmp=(unsigned char *)malloc(realWidth*realHeight*4);

            if (!compressed)
            {
                fread(tmp,4,realWidth*realHeight,fp);
                ret=true;
            }
            else
            {
                unsigned char *compr=(unsigned char *)malloc(realWidth*realHeight*4);
                fread(compr,1,realWidth*realHeight*4,fp);
                tmp2=(unsigned char *)malloc(realWidth*realHeight*4);
                RLEUncompress(tmp2,compr,realWidth*realHeight*4);
                for (t=0;t<realWidth*realHeight;t++)
                {
                    tmp[t*4]=tmp2[t+realWidth*realHeight*0];
                    tmp[t*4+1]=tmp2[t+realWidth*realHeight*1];
                    tmp[t*4+2]=tmp2[t+realWidth*realHeight*2];
                    tmp[t*4+3]=tmp2[t+realWidth*realHeight*3];
                }

                free((void *)compr);
                free((void *)tmp2);
                ret=true;
            }

            for (y=0;y<texHeight;y++)
                for (x=0;x<texWidth;x++)
                {
                    int d=y*texWidth+x;
                    if (x<realWidth && y<realHeight)
                    {
                        int c=y*realWidth+x;
                        img[d].red=tmp[c*4];
                        img[d].green=tmp[c*4+1];
                        img[d].blue=tmp[c*4+2];
                        img[d].alpha=tmp[c*4+3];
                    }
                    else
                    {
                        img[d].red=img[d].green=img[d].blue=img[d].alpha=0;
                    }
                }
            free((void *)tmp);
        }
        fclose(fp);
    }
    else
    {
        printf("[Image::loadRGBA] Error: %s not found\n",filename);
    }
    return true;
}

void Image::blit(int dx,int dy, Image *source, int sx, int sy, int sw, int sh)
{
    ColorRGBA t;
    int c,d;
    for (c=0;c<sw;c++)
        for (d=0;d<sh;d++)
        {
            t=source->getPixel(sx+c,sy+d);
            setPixel(dx+c,dy+d,t);
        }
}

int Image::alphaCollision(int dx,int dy, Image *source, int sx, int sy, int sw, int sh, int alphaMinimum)
{
    ColorRGBA t1, t2;
    int r=0;
    int c,d;
    for (c=0;c<sw;c++)
        for (d=0;d<sh;d++)
        {
            t1=source->getPixel(sx+c,sy+d);
            t2=getPixel(dx+sx+c,dy+sy+d);
            if (t1.alpha>alphaMinimum && t2.alpha>alphaMinimum)
                r++;
        }
    return r;
}
